#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import ValueType, KeyField, ListField, StringField
from ..utils import broadcast_shape, check_out_param
from ..operands import TensorOperand, TensorOperandMixin
from ..datasource import tensor as astensor
from ..core import Tensor, TensorOrder
from ..array_utils import as_same_device, device


class TensorChoose(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.CHOOSE

    _a = KeyField('a')
    _choices = ListField('choices', ValueType.key)
    _mode = StringField('mode')

    def __init__(self, mode=None, dtype=None, **kw):
        super().__init__(_mode=mode, _dtype=dtype, **kw)

    def __setattr__(self, key, value):
        if key == '_mode' and value not in ('raise', 'wrap', 'clip'):
            raise ValueError('mode should be raise, wrap or clip')

        super().__setattr__(key, value)

    @property
    def a(self):
        return self._a

    @property
    def choices(self):
        return self._choices

    @property
    def mode(self):
        return self._mode

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._a = self._inputs[0]
        self._choices = self._inputs[1:]

    def __call__(self, a, choices, out=None):
        if out is not None and not isinstance(out, Tensor):
            raise TypeError(f'out should be Tensor object, got {type(out)} instead')

        inputs = [a] + choices
        shape = broadcast_shape(a.shape, *[c.shape for c in choices])
        order = TensorOrder.C_ORDER if out is None else out.order
        t = self.new_tensor(inputs, shape, order=order)

        if out is None:
            return t

        check_out_param(out, t, 'unsafe')
        out_shape, out_dtype = out.shape, out.dtype
        # if `out` is specified, use out's dtype and shape
        if out_shape != t.shape:
            raise ValueError(f'output shape should be {t.shape}, got {out_shape}')
        setattr(self, '_dtype', out_dtype)
        out.data = t.data
        return out

    @classmethod
    def tile(cls, op):
        from ..arithmetic.core import TensorElementWise

        return TensorElementWise.tile(op)

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)
        a, choices = inputs[0], inputs[1:]

        out = op.outputs[0]
        with device(device_id):
            ctx[out.key] = xp.choose(a, choices, mode=op.mode).astype(
                op.dtype, order=out.order.value, copy=False)


def choose(a, choices, out=None, mode='raise'):
    """
    Construct a tensor from an index tensor and a set of tensors to choose from.

    First of all, if confused or uncertain, definitely look at the Examples -
    in its full generality, this function is less simple than it might
    seem from the following code description (below ndi =
    `mt.lib.index_tricks`):

    ``mt.choose(a,c) == mt.array([c[a[I]][I] for I in ndi.ndindex(a.shape)])``.

    But this omits some subtleties.  Here is a fully general summary:

    Given an "index" tensor (`a`) of integers and a sequence of `n` tensors
    (`choices`), `a` and each choice tensor are first broadcast, as necessary,
    to tensors of a common shape; calling these *Ba* and *Bchoices[i], i =
    0,...,n-1* we have that, necessarily, ``Ba.shape == Bchoices[i].shape``
    for each `i`.  Then, a new array with shape ``Ba.shape`` is created as
    follows:

    * if ``mode=raise`` (the default), then, first of all, each element of
      `a` (and thus `Ba`) must be in the range `[0, n-1]`; now, suppose that
      `i` (in that range) is the value at the `(j0, j1, ..., jm)` position
      in `Ba` - then the value at the same position in the new array is the
      value in `Bchoices[i]` at that same position;

    * if ``mode=wrap``, values in `a` (and thus `Ba`) may be any (signed)
      integer; modular arithmetic is used to map integers outside the range
      `[0, n-1]` back into that range; and then the new array is constructed
      as above;

    * if ``mode=clip``, values in `a` (and thus `Ba`) may be any (signed)
      integer; negative integers are mapped to 0; values greater than `n-1`
      are mapped to `n-1`; and then the new tensor is constructed as above.

    Parameters
    ----------
    a : int tensor
        This tensor must contain integers in `[0, n-1]`, where `n` is the number
        of choices, unless ``mode=wrap`` or ``mode=clip``, in which cases any
        integers are permissible.
    choices : sequence of tensors
        Choice tensors. `a` and all of the choices must be broadcastable to the
        same shape.  If `choices` is itself a tensor (not recommended), then
        its outermost dimension (i.e., the one corresponding to
        ``choices.shape[0]``) is taken as defining the "sequence".
    out : tensor, optional
        If provided, the result will be inserted into this tensor. It should
        be of the appropriate shape and dtype.
    mode : {'raise' (default), 'wrap', 'clip'}, optional
        Specifies how indices outside `[0, n-1]` will be treated:

          * 'raise' : an exception is raised
          * 'wrap' : value becomes value mod `n`
          * 'clip' : values < 0 are mapped to 0, values > n-1 are mapped to n-1

    Returns
    -------
    merged_array : Tensor
        The merged result.

    Raises
    ------
    ValueError: shape mismatch
        If `a` and each choice tensor are not all broadcastable to the same
        shape.

    See Also
    --------
    Tensor.choose : equivalent method

    Notes
    -----
    To reduce the chance of misinterpretation, even though the following
    "abuse" is nominally supported, `choices` should neither be, nor be
    thought of as, a single tensor, i.e., the outermost sequence-like container
    should be either a list or a tuple.

    Examples
    --------

    >>> import mars.tensor as mt

    >>> choices = [[0, 1, 2, 3], [10, 11, 12, 13],
    ...   [20, 21, 22, 23], [30, 31, 32, 33]]
    >>> mt.choose([2, 3, 1, 0], choices
    ... # the first element of the result will be the first element of the
    ... # third (2+1) "array" in choices, namely, 20; the second element
    ... # will be the second element of the fourth (3+1) choice array, i.e.,
    ... # 31, etc.
    ... ).execute()
    array([20, 31, 12,  3])
    >>> mt.choose([2, 4, 1, 0], choices, mode='clip').execute() # 4 goes to 3 (4-1)
    array([20, 31, 12,  3])
    >>> # because there are 4 choice arrays
    >>> mt.choose([2, 4, 1, 0], choices, mode='wrap').execute() # 4 goes to (4 mod 4)
    array([20,  1, 12,  3])
    >>> # i.e., 0

    A couple examples illustrating how choose broadcasts:

    >>> a = [[1, 0, 1], [0, 1, 0], [1, 0, 1]]
    >>> choices = [-10, 10]
    >>> mt.choose(a, choices).execute()
    array([[ 10, -10,  10],
           [-10,  10, -10],
           [ 10, -10,  10]])

    >>> # With thanks to Anne Archibald
    >>> a = mt.array([0, 1]).reshape((2,1,1))
    >>> c1 = mt.array([1, 2, 3]).reshape((1,3,1))
    >>> c2 = mt.array([-1, -2, -3, -4, -5]).reshape((1,1,5))
    >>> mt.choose(a, (c1, c2)).execute() # result is 2x3x5, res[0,:,:]=c1, res[1,:,:]=c2
    array([[[ 1,  1,  1,  1,  1],
            [ 2,  2,  2,  2,  2],
            [ 3,  3,  3,  3,  3]],
           [[-1, -2, -3, -4, -5],
            [-1, -2, -3, -4, -5],
            [-1, -2, -3, -4, -5]]])

    """
    a = astensor(a, dtype='i8')
    choices = [astensor(c) for c in choices]

    dtype = np.result_type(*[c.dtype for c in choices])
    op = TensorChoose(mode=mode, dtype=dtype)
    return op(a, choices, out=out)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import numpy as np

from ... import opcodes as OperandDef
from ...config import options
from ...context import get_context, RunningMode
from ...serialize import TupleField, ListField, KeyField, StringField
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape
from ..datasource import tensor as astensor
from .core import TensorDataStore

try:
    import vineyard
except ImportError:
    vineyard = None


# Note [Tensor to vineyard]
#
#   step 1: store local chunks
#   step 2: run iterative tiling, to gather, and build local blob of chunks id, with `expect_worker`
#   step 3: build chunk_map

class TensorVineyardDataStoreChunk(TensorDataStore):
    _op_type_ = OperandDef.TENSOR_STORE_VINEYARD_CHUNK

    _input = KeyField('input')

    # vineyard ipc socket
    _vineyard_socket = StringField('vineyard_socket')

    def __init__(self, vineyard_socket=None, dtype=None, sparse=None, **kw):
        super().__init__(_vineyard_socket=vineyard_socket, _dtype=dtype, _sparse=sparse, **kw)

    @property
    def vineyard_socket(self):
        return self._vineyard_socket

    @classmethod
    def _get_out_chunk(cls, op, in_chunk):
        chunk_op = op.copy().reset_key()
        out_chunk_shape = (np.nan,) * in_chunk.ndim
        return chunk_op.new_chunk([in_chunk], shape=out_chunk_shape,
                                  index=in_chunk.index)

    @classmethod
    def tile(cls, op):
        # Not truely necessary, but putting a barrier here will make things simple
        check_chunks_unknown_shape(op.inputs, TilesError)

        out = super().tile(op)[0]
        new_op = op.copy().reset_key()
        return new_op.new_tensors(out.inputs, shape=op.input.shape, dtype=out.dtype,
                                  chunks=out.chunks, nsplits=((np.nan,),))

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError('vineyard is not available')
        client = vineyard.connect(op.vineyard_socket)

        np_value = ctx[op.input.key]
        if not np_value.flags['C_CONTIGUOUS']:
            np_value = np.ascontiguousarray(np_value)
        tensor = vineyard.TensorBuilder.from_numpy(client, np_value)
        tensor.shape = op.input.shape
        tensor.chunk_index = op.input.index
        tensor = tensor.build(client)
        tensor.persist(client)

        # store the result object id to execution context
        ctx[op.outputs[0].key] = np.array([(client.instance_id, tensor.id)])


class TensorVineyardDataStoreChunkMap(TensorDataStore):
    _op_type_ = OperandDef.TENSOR_STORE_VINEYARD_CHUNK_MAP

    _input = KeyField('input')
    _local_chunks = ListField("local_chunks")

    # vineyard ipc socket
    _vineyard_socket = StringField('vineyard_socket')

    def __init__(self, vineyard_socket=None, dtype=None, sparse=None, **kw):
        super().__init__(_vineyard_socket=vineyard_socket, _dtype=dtype, _sparse=sparse, **kw)

    @property
    def local_chunks(self):
        return self._local_chunks

    @property
    def vineyard_socket(self):
        return self._vineyard_socket

    @classmethod
    def _process_out_chunks(cls, op, out_chunks):
        merge_op = TensorVineyardDataStoreGlobalMeta(
                vineyard_socket=op.vineyard_socket,
                chunk_shape=op.inputs[0].chunk_shape, shape=op.inputs[0].shape,
                sparse=op.sparse, dtype=op.inputs[0].dtype)
        return merge_op.new_chunks(out_chunks, shape=out_chunks[0].shape,
                                   index=(0,) * out_chunks[0].ndim)

    @classmethod
    def tile(cls, op):
        check_chunks_unknown_shape(op.inputs, TilesError)

        if vineyard is None:
            raise RuntimeError('vineyard is not available')
        client = vineyard.connect(op.vineyard_socket)

        ctx = get_context()
        if ctx.running_mode == RunningMode.distributed:
            metas = ctx.get_worker_metas()
            workers = {meta['vineyard']['instance_id']: addr for addr, meta in metas.items()}
        else:
            workers = {client.instance_id: '127.0.0.1'}

        all_chunk_ids = np.concatenate(ctx.get_chunk_results([c.key for c in op.input.chunks]))
        chunk_map = defaultdict(list)
        for instance_id, chunk_id in all_chunk_ids:
            chunk_map[instance_id].append(chunk_id)

        out_chunks = []
        for idx, (instance_id, local_chunks) in enumerate(chunk_map.items()):
            chunk_op = op.copy().reset_key()
            chunk_op._local_chunks = local_chunks
            chunk_op._expect_worker = workers[instance_id]
            out_chunks.append(chunk_op.new_chunk(op.input.chunks, shape=(1,), index=(idx,)))
        out_chunks = cls._process_out_chunks(op, out_chunks)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=(len(out_chunks),), chunks=out_chunks,
                                  nsplits=((1,) * len(out_chunks),))

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError('vineyard is not available')
        client = vineyard.connect(op.vineyard_socket)
        chunk_blob = vineyard.UInt64VectorBuilder(client, op.local_chunks).build(client)
        ctx[op.outputs[0].key] = (client.instance_id, chunk_blob.id)


class TensorVineyardDataStoreGlobalMeta(TensorDataStore):
    _op_type_ = OperandDef.TENSOR_STORE_VINEYARD_GLOBAL_META

    _input = KeyField('input')
    _shape = TupleField('shape')
    _chunk_shape = TupleField('chunk_shape')

    # vineyard ipc socket
    _vineyard_socket = StringField('vineyard_socket')

    def __init__(self, vineyard_socket=None,
                 chunk_shape=None, shape=None, dtype=None, sparse=None, **kw):
        super().__init__(_vineyard_socket=vineyard_socket,
                         _chunk_shape=chunk_shape, _shape=shape, _dtype=dtype, _sparse=sparse, **kw)

    @property
    def shape(self):
        return self._shape

    @property
    def chunk_shape(self):
        return self._chunk_shape

    @property
    def vineyard_socket(self):
        return self._vineyard_socket

    @classmethod
    def _process_out_chunks(cls, op, out_chunks):
        if len(out_chunks) == 1:
            return out_chunks
        else:
            raise NotImplementedError('not implemented')

    @classmethod
    def tile(cls, op):
        return [super().tile(op)[0]]

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError('vineyard is not available')
        client = vineyard.connect(op.vineyard_socket)

        tensor = vineyard.GlobalTensorBuilder(client)
        for in_chunk in op.inputs:
            instance_id, chunk_blob_id = ctx[in_chunk.key]
            tensor.add_chunks(instance_id, chunk_blob_id)
        tensor.shape = op.shape
        tensor.chunk_shape = op.chunk_shape
        tensor = tensor.build(client)
        tensor.persist(client)

        # store the result object id to execution context
        ctx[op.outputs[0].key] = vineyard.ObjectID.stringify(tensor.id)


def tovineyard(x, vineyard_socket=None):
    if vineyard_socket is None:
        vineyard_socket = options.vineyard.socket

    x = astensor(x)
    chunk_op = TensorVineyardDataStoreChunk(vineyard_socket=vineyard_socket,
                                            dtype=x.dtype, sparse=x.issparse())
    chunk_map_op = TensorVineyardDataStoreChunkMap(vineyard_socket=vineyard_socket,
                                                   dtype=x.dtype, sparse=x.issparse())
    return chunk_map_op(chunk_op(x))



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple, OrderedDict
from collections.abc import Iterable
from typing import Dict

import numpy as np
import pandas as pd

from ...operands import OperandStage
from ...serialize import AnyField, BoolField, Int32Field, Int64Field, \
    DictField, StringField
from ...utils import tokenize
from ..core import DATAFRAME_TYPE
from ..merge import DataFrameConcat
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_df, parse_index, build_empty_series, \
    filter_dtypes_by_index

_stage_info = namedtuple('_stage_info', ('map_groups', 'map_sources', 'combine_sources',
                                         'combine_columns', 'combine_funcs', 'key_to_funcs',
                                         'valid_columns', 'min_periods_func_name'))


class BaseDataFrameExpandingAgg(DataFrameOperand, DataFrameOperandMixin):
    _min_periods = Int64Field('min_periods')
    _axis = Int32Field('axis')
    _func = AnyField('func')

    # always treat count as valid. this behavior is cancelled in pandas 1.0
    _count_always_valid = BoolField('count_always_valid')
    # True if function name is treated as new index
    _append_index = BoolField('append_index')

    # chunk params
    _output_agg = BoolField('output_agg')

    _map_groups = DictField('map_groups')
    _map_sources = DictField('map_sources')
    _combine_sources = DictField('combine_sources')
    _combine_columns = DictField('combine_columns')
    _combine_funcs = DictField('combine_funcs')
    _key_to_funcs = DictField('keys_to_funcs')

    _min_periods_func_name = StringField('min_periods_func_name')

    def __init__(self, min_periods=None, axis=None, func=None, count_always_valid=None,
                 append_index=None, output_agg=False, map_groups=None, map_sources=None,
                 combine_sources=None, combine_columns=None, combine_funcs=None,
                 key_to_funcs=None, min_periods_func_name=None, stage=None, **kw):
        super().__init__(_min_periods=min_periods, _axis=axis, _func=func, _stage=stage,
                         _count_always_valid=count_always_valid, _append_index=append_index,
                         _output_agg=output_agg, _map_groups=map_groups, _map_sources=map_sources,
                         _combine_sources=combine_sources, _combine_columns=combine_columns,
                         _combine_funcs=combine_funcs, _key_to_funcs=key_to_funcs,
                         _min_periods_func_name=min_periods_func_name, **kw)

    @property
    def min_periods(self) -> int:
        return self._min_periods

    @property
    def axis(self) -> int:
        return self._axis

    @property
    def func(self):
        return self._func

    @property
    def count_always_valid(self):
        return self._count_always_valid

    @property
    def append_index(self):
        return self._append_index

    @property
    def output_agg(self):
        return self._output_agg

    @property
    def map_groups(self) -> Dict:
        return self._map_groups

    @property
    def map_sources(self) -> Dict:
        return self._map_sources

    @property
    def combine_sources(self) -> Dict:
        return self._combine_sources

    @property
    def combine_columns(self) -> Dict:
        return self._combine_columns

    @property
    def combine_funcs(self) -> Dict:
        return self._combine_funcs

    @property
    def key_to_funcs(self) -> Dict:
        return self._key_to_funcs

    @property
    def min_periods_func_name(self) -> str:
        return self._min_periods_func_name

    @property
    def output_limit(self):
        return 2 if self.output_agg else 1

    def __call__(self, expanding):
        inp = expanding.input
        raw_func = self.func
        self._normalize_funcs()

        if isinstance(inp, DATAFRAME_TYPE):
            empty_df = build_df(inp)
            for c, t in empty_df.dtypes.items():
                if t == np.dtype('O'):
                    empty_df[c] = 'O'

            test_df = expanding(empty_df).agg(raw_func)
            if self._axis == 0:
                index_value = inp.index_value
            else:
                index_value = parse_index(test_df.index,
                                          expanding.params, inp,
                                          store_data=False)
            self._append_index = test_df.columns.nlevels != empty_df.columns.nlevels
            return self.new_dataframe(
                [inp], shape=(inp.shape[0], test_df.shape[1]),
                dtypes=test_df.dtypes, index_value=index_value,
                columns_value=parse_index(test_df.columns, store_data=True))
        else:
            pd_index = inp.index_value.to_pandas()
            empty_series = build_empty_series(inp.dtype, index=pd_index[:0], name=inp.name)
            test_obj = expanding(empty_series).agg(raw_func)
            if isinstance(test_obj, pd.DataFrame):
                return self.new_dataframe([inp], shape=(inp.shape[0], test_obj.shape[1]),
                                          dtypes=test_obj.dtypes, index_value=inp.index_value,
                                          columns_value=parse_index(test_obj.dtypes.index, store_data=True))
            else:
                return self.new_series([inp], shape=inp.shape, dtype=test_obj.dtype,
                                       index_value=inp.index_value, name=test_obj.name)

    def _normalize_funcs(self):
        if isinstance(self._func, dict):
            new_func = OrderedDict()
            for k, v in self._func.items():
                if isinstance(v, str) or callable(v):
                    new_func[k] = [v]
                else:
                    new_func[k] = v
            self._func = new_func
        elif isinstance(self._func, Iterable) and not isinstance(self._func, str):
            self._func = list(self._func)
        else:
            self._func = [self._func]

    @staticmethod
    def _safe_append(d, key, val):
        if key not in d:
            d[key] = []
        if val not in d[key]:
            d[key].append(val)

    @classmethod
    def _get_stage_functions(cls, op: "BaseDataFrameExpandingAgg", func):
        raise NotImplementedError

    @classmethod
    def _gen_chunk_stage_info(cls, op: "BaseDataFrameExpandingAgg", chunk_cols=None, min_periods=1):
        map_groups = OrderedDict()
        map_sources = OrderedDict()
        combine_sources = OrderedDict()
        combine_columns = OrderedDict()
        combine_funcs = OrderedDict()
        key_to_funcs = OrderedDict()
        valid_columns = []
        min_periods_func_name = None

        def _clean_dict(d):
            return OrderedDict((k, sorted(v) if v != [None] else None) for k, v in d.items())

        def _fun_to_str(fun):
            if isinstance(fun, str):
                return fun
            fun_str = tokenize(fun)
            key_to_funcs[fun_str] = fun
            return fun if isinstance(fun, str) else tokenize(fun)

        def _add_column_to_functions(col, fun_name, mappers, aggregator):
            sources = []
            for mapper in mappers:
                mapper_str = _fun_to_str(mapper)
                cls._safe_append(map_groups, mapper_str, col)
                sources.append(mapper_str)

            combine_sources[fun_name] = sources
            cls._safe_append(combine_columns, fun_name, col)
            combine_funcs[fun_name] = _fun_to_str(aggregator)

        chunk_cols = set(chunk_cols) if chunk_cols is not None else None
        if isinstance(op.func, list):
            op_func = {None: op.func}
        else:
            op_func = op.func

        for col, funcs in op_func.items():
            if col is not None:
                if chunk_cols is not None and col not in chunk_cols:
                    continue
                valid_columns.append(col)

            if min_periods > 1:
                min_periods_func_name = tokenize(chunk_cols, 'min_periods')
                _add_column_to_functions(col, min_periods_func_name,
                                         *cls._get_stage_functions(op, '_data_count'))

            for func in funcs:
                mapper_funcs, combine_func = cls._get_stage_functions(op, func)
                _add_column_to_functions(col, func, mapper_funcs, combine_func)

        return _stage_info(map_groups=_clean_dict(map_groups), map_sources=map_sources,
                           combine_sources=combine_sources, combine_columns=_clean_dict(combine_columns),
                           combine_funcs=combine_funcs, key_to_funcs=key_to_funcs,
                           valid_columns=valid_columns or None, min_periods_func_name=min_periods_func_name)

    @classmethod
    def _remap_dtypes(cls, in_df, out_df):
        if in_df.ndim == 1:
            if out_df.ndim == 2:
                return {0: (0, out_df.dtypes)}, (in_df.nsplits[0], (len(out_df.dtypes),))
            return None, in_df.nsplits

        axis = out_df.op.axis
        chunk_idx_to_dtypes = dict()
        new_dtypes_sizes = []
        for c in in_df.cix[0, :]:
            columns = c.columns_value.to_pandas()
            dtypes = filter_dtypes_by_index(out_df.dtypes, columns)

            if len(dtypes):
                chunk_idx_to_dtypes[c.index[1]] = (len(chunk_idx_to_dtypes), dtypes)
                new_dtypes_sizes.append(len(dtypes))
        new_nsplits = list(in_df.nsplits)
        new_nsplits[1 - axis] = tuple(new_dtypes_sizes)
        return chunk_idx_to_dtypes, tuple(new_nsplits)

    @classmethod
    def _tile_single(cls, op: "BaseDataFrameExpandingAgg"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        chunk_idx_to_dtypes, new_nsplits = cls._remap_dtypes(in_df, out_df)

        chunks = []
        for c in in_df.chunks:
            try:
                if out_df.ndim == 2:
                    new_axis_idx, new_dtypes = chunk_idx_to_dtypes[c.index[1] if c.ndim > 1 else 0]
                else:
                    new_axis_idx, new_dtypes = None, None
            except KeyError:
                continue

            chunk_op = op.copy().reset_key()

            if out_df.ndim == 2:
                chunks.append(chunk_op.new_chunk(
                    [in_df.chunks[0]], dtypes=new_dtypes, index=(c.index[0], new_axis_idx),
                    shape=(c.shape[0], len(new_dtypes)), index_value=c.index_value,
                    columns_value=parse_index(new_dtypes.index, store_data=True)))
            else:
                params = c.params.copy()
                params['dtype'] = out_df.dtype
                chunks.append(chunk_op.new_chunk([in_df.chunks[0]], **params))

        tileable_op = op.copy().reset_key()
        params = out_df.params.copy()
        params['chunks'] = chunks
        if new_nsplits:
            params['nsplits'] = new_nsplits
        return tileable_op.new_tileables([in_df], **params)

    @classmethod
    def tile(cls, op: "BaseDataFrameExpandingAgg"):
        axis = op.axis

        in_df = op.inputs[0]
        out_df = op.outputs[0]

        if in_df.chunk_shape[op.axis] == 1:
            return cls._tile_single(op)

        dtypes_mapping, new_nsplits = cls._remap_dtypes(in_df, out_df)
        new_chunk_shape = tuple(len(split) for split in new_nsplits)

        data_chunks = []
        summary_chunks = np.empty(new_chunk_shape, dtype=np.object)
        stage_info_dict = dict()
        for c in in_df.chunks:
            try:
                if out_df.ndim == 2:
                    new_axis_idx, new_dtypes = dtypes_mapping[c.index[1] if c.ndim > 1 else 0]
                else:
                    new_axis_idx, new_dtypes = None, None
            except KeyError:
                continue

            new_index = (c.index[0], new_axis_idx)

            try:
                stage_info = stage_info_dict[new_index[1]]
            except KeyError:
                cols = c.dtypes.index if c.ndim == 2 else None
                stage_info = stage_info_dict[new_index[1]] \
                    = cls._gen_chunk_stage_info(op, cols, min_periods=op.min_periods)

            chunk_op = op.copy().reset_key()
            chunk_op._output_agg = c.index[axis] != in_df.chunk_shape[axis] - 1
            chunk_op._stage = OperandStage.map
            chunk_op._map_sources = stage_info.map_sources
            chunk_op._map_groups = stage_info.map_groups
            chunk_op._key_to_funcs = stage_info.key_to_funcs

            if out_df.ndim == 2:
                kw0 = dict(dtypes=new_dtypes, index=new_index,
                           shape=(c.shape[0], len(new_dtypes)),
                           index_value=c.index_value,
                           columns_value=parse_index(new_dtypes.index, store_data=True))
                kw1 = kw0.copy()
                kw1['shape'] = (1, len(new_dtypes)) if axis == 0 else (c.shape[0], 1)
            else:
                kw0 = dict(dtype=out_df.dtype, index=c.index, shape=c.shape,
                           name=c.name, index_value=c.index_value)
                kw1 = kw0.copy()
                kw1['shape'] = (1,)
            out_chunks = chunk_op.new_chunks([c], [kw0, kw1])
            data_chunks.append(out_chunks[0])
            if chunk_op.output_agg:
                summary_chunks[new_index] = out_chunks[1]

        chunks = []
        for c in data_chunks:
            stage_info = stage_info_dict[c.index[1] if c.ndim > 1 else None]

            chunk_op = op.copy().reset_key()
            chunk_op._output_agg = False
            chunk_op._stage = OperandStage.combine
            chunk_op._map_groups = stage_info.map_groups
            chunk_op._combine_sources = stage_info.combine_sources
            chunk_op._combine_columns = stage_info.combine_columns
            chunk_op._combine_funcs = stage_info.combine_funcs
            chunk_op._key_to_funcs = stage_info.key_to_funcs
            chunk_op._min_periods_func_name = stage_info.min_periods_func_name

            params = c.params.copy()
            if c.ndim == 2:
                summary_inputs = list(summary_chunks[:c.index[0], c.index[1]])
            else:
                summary_inputs = list(summary_chunks[:c.index[0]])

            if len(summary_inputs) > 1:
                concat_op = DataFrameConcat(output_types=out_df.op.output_types, axis=op.axis)
                concat_summary = concat_op.new_chunk(summary_inputs)
                chunks.append(chunk_op.new_chunk([c, concat_summary], **params))
            elif len(summary_inputs) == 1:
                chunks.append(chunk_op.new_chunk([c, summary_inputs[0]], **params))
            else:
                chunks.append(chunk_op.new_chunk([c], **params))

        df_op = op.copy().reset_key()
        params = out_df.params.copy()
        params.update(dict(chunks=chunks, nsplits=new_nsplits))
        return df_op.new_tileables([in_df], **params)

    @classmethod
    def _execute_map_function(cls, op: "BaseDataFrameExpandingAgg", func, in_data):
        raise NotImplementedError

    @classmethod
    def _execute_map(cls, ctx, op: "BaseDataFrameExpandingAgg"):
        in_data = ctx[op.inputs[0].key]

        # map according to map groups
        map_results = []
        summary_results = []
        for map_func_str, cols in op.map_groups.items():
            if cols is None:
                src_df = in_data
            else:
                src_df = in_data[cols]

            result, summary = cls._execute_map_function(op, map_func_str, src_df)
            map_results.append(result)
            if op.output_agg:
                summary_results.append(summary)

        if op.output_agg:
            summary_results.append(pd.Series([len(in_data)], index=summary_results[0].index))

        ctx[op.outputs[0].key] = tuple(map_results)
        if op.output_agg:
            ctx[op.outputs[1].key] = tuple(summary_results)

    @classmethod
    def _append_func_name_index(cls, op: "BaseDataFrameExpandingAgg", df, func_name):
        if not op.append_index:
            return

        col_frame = df.columns.to_frame().copy()
        col_frame[len(col_frame.columns)] = func_name
        df.columns = pd.MultiIndex.from_frame(
            col_frame, names=tuple(df.columns.names) + (None,))

    @classmethod
    def _execute_combine_function(cls, op: "BaseDataFrameExpandingAgg", func,
                                  pred_inputs, local_inputs, func_cols):
        raise NotImplementedError

    @classmethod
    def _execute_combine(cls, ctx, op: "BaseDataFrameExpandingAgg"):
        out_df = op.outputs[0]
        local_data = ctx[op.inputs[0].key]
        local_data_dict = dict(zip(op.map_groups.keys(), local_data))

        func_to_aggs = OrderedDict()

        if len(op.inputs) == 1:
            pred_record_count = 0
            for func_name, func_sources in op.combine_sources.items():
                func_str = op.combine_funcs[func_name]
                func_cols = op.combine_columns[func_name]
                if func_cols is None:
                    local_inputs = [local_data_dict[src] for src in func_sources]
                else:
                    local_inputs = [local_data_dict[src][func_cols] for src in func_sources]

                func = op.key_to_funcs[func_str]
                func_to_aggs[func_name] = cls._execute_combine_function(
                    op, func, None, local_inputs, func_cols)
        else:
            pred_data = ctx[op.inputs[1].key]
            pred_record_count = pred_data[-1].sum()
            pred_data_dict = dict(zip(op.map_groups.keys(), pred_data))

            for func_name, func_sources in op.combine_sources.items():
                func_str = op.combine_funcs[func_name]
                func_cols = op.combine_columns[func_name]
                if func_cols is None:
                    local_inputs = [local_data_dict[src] for src in func_sources]
                    pred_inputs = [pred_data_dict[src] for src in func_sources]
                else:
                    local_inputs = [local_data_dict[src][func_cols] for src in func_sources]
                    pred_inputs = [pred_data_dict[src][func_cols] for src in func_sources]

                func = op.key_to_funcs[func_str]
                func_to_aggs[func_name] = cls._execute_combine_function(
                    op, func, pred_inputs, local_inputs, func_cols)

        if op.min_periods_func_name is not None:
            valid_counts = func_to_aggs.pop(op.min_periods_func_name)
            invalid_poses = valid_counts < op.min_periods
            for func_name in func_to_aggs.keys():
                if func_name == 'count':
                    if not op.count_always_valid and pred_record_count < op.min_periods - 1:
                        func_to_aggs[func_name].iloc[:op.min_periods - pred_record_count - 1] = np.nan
                else:
                    func_to_aggs[func_name][invalid_poses] = np.nan

        for func_name, agg_df in func_to_aggs.items():
            if out_df.ndim == 2 and agg_df.ndim == 1:
                agg_df.name = func_name
                agg_df = func_to_aggs[func_name] = pd.DataFrame(agg_df)
            cls._append_func_name_index(op, agg_df, func_name)

        if len(func_to_aggs) == 1:
            val = list(func_to_aggs.values())[0]
        else:
            out_df = op.outputs[0]
            val = pd.concat(list(func_to_aggs.values()), axis=1 - op.axis)

        if out_df.ndim > 1:
            val = val.reindex(out_df.columns_value.to_pandas(), axis=1 - op.axis, copy=False)
        else:
            val.name = out_df.name
        ctx[op.outputs[0].key] = val

    @classmethod
    def _execute_raw_function(cls, op: "BaseDataFrameExpandingAgg", in_data):
        raise NotImplementedError

    @classmethod
    def execute(cls, ctx, op: "BaseDataFrameExpandingAgg"):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        elif op.stage == OperandStage.combine:
            cls._execute_combine(ctx, op)
        else:
            in_data = ctx[op.inputs[0].key]
            out_df = op.outputs[0]

            r = cls._execute_raw_function(op, in_data)
            if out_df.ndim == 1:
                r = r.iloc[:, 0]
                r.name = out_df.name
            ctx[op.outputs[0].key] = r



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from abc import ABC, abstractmethod
from collections import OrderedDict, namedtuple
from enum import Enum
from operator import itemgetter
from typing import Tuple, List, Union
from numbers import Integral

import numpy as np

from ...core import TileableEntity
from ...operands import OperandStage
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape, calc_nsplits, \
    merge_chunks, recursive_tile
from ..core import TENSOR_TYPE, Chunk, TensorOrder
from ..operands import TensorShuffleProxy
from ..utils import slice_split, calc_sliced_size, broadcast_shape, unify_chunks, \
    split_indexes_into_chunks, filter_inputs, calc_pos


class IndexType(Enum):
    new_axis = 0
    slice = 1
    label_slice = 2  # e.g. 'a': 'd' used for pandas etc
    integer = 3
    label = 4  # e.g. 'a' used for pandas etc
    bool_index = 5
    fancy_index = 6
    label_fancy_index = 7  # e.g. ['a', 'b', 'c'] for pandas etc


class IndexInfo:
    def __init__(self,
                 index_type: IndexType,
                 input_axis: int,
                 output_axis: int,
                 raw_index,
                 handler):
        self.index_type = index_type
        self.input_axis = input_axis
        self.output_axis = output_axis
        self.raw_index = raw_index
        self.handler = handler


class FancyIndexInfo(IndexInfo):
    def __init__(self,
                 index_type: IndexType,
                 input_axis: int,
                 output_axis: int,
                 raw_index,
                 handler):
        super().__init__(index_type, input_axis, output_axis,
                         raw_index, handler)

        # extra info for fancy index
        # shape broadcast index
        self.shape_unified_index = None
        # split info
        #   - chunk_index_to_fancy_index_arrays
        #   - chunk_index_to_raw_positions
        #   - is_fancy_index_asc_sorted
        self.split_info = None


ChunkIndexAxisInfo = namedtuple(
    'chunk_index_axis_info',
    ['output_axis_index', 'processed_index', 'output_shape'])


class ChunkIndexInfo:
    def __init__(self):
        self.indexes = []
        self.output_chunk_index = []
        self.output_chunk_shape = []

    def set(self, info: ChunkIndexAxisInfo):
        output_axis_index = info.output_axis_index
        if output_axis_index is not None:
            self.output_chunk_index.append(output_axis_index)
        self.indexes.append(info.processed_index)
        output_shape = info.output_shape
        if output_shape is not None:
            if not isinstance(output_shape, tuple):
                self.output_chunk_shape.append(output_shape)
            else:
                self.output_chunk_shape.extend(output_shape)


class IndexHandlerContext(ABC):
    def __init__(self, op):
        self.parsed_infos = []
        self.input_axis = 0
        self.output_axis = 0

        # store index_type -> positions
        # for a quick search on indexes of a specified index type
        self._index_type_to_positions = dict()

        # store chunk index -> ChunkIndexInfo
        # for the IndexHandler to process
        self.chunk_index_to_info = OrderedDict()
        self.op = op
        self.tileable = op.input
        self.set_tileable(self.tileable)

        # chunks and nsplits, used for store intermediate result
        self.processed_chunks = None
        self.out_chunks = None
        self.out_nsplits = None

    def append(self, index_info: IndexInfo):
        position = len(self.parsed_infos)
        if index_info.index_type not in self._index_type_to_positions:
            self._index_type_to_positions[index_info.index_type] = []
        self._index_type_to_positions[index_info.index_type].append(position)
        self.parsed_infos.append(index_info)

    def get_positions(self, index_type: IndexType) -> List[int]:
        return self._index_type_to_positions.get(index_type, [])

    def get_indexes(self, index_type: IndexType):
        return [self.parsed_infos[i] for i in self.get_positions(index_type)]

    def set_tileable(self, tileable: TileableEntity):
        for chunk in tileable.chunks:
            self.chunk_index_to_info[chunk.index] = ChunkIndexInfo()

    @abstractmethod
    def concat_chunks(self,
                      chunks: List[Chunk],
                      axis: Union[Tuple, int]) -> Chunk:
        pass

    @abstractmethod
    def create_chunk(self,
                     chunk_index: Tuple[int],
                     chunk_index_info: ChunkIndexInfo) -> Chunk:
        pass

    def create_tileable(self) -> TileableEntity:
        out = self.op.outputs[0]
        params = out.params
        params['chunks'] = self.out_chunks
        params['nsplits'] = self.out_nsplits
        if 'shape' in params and any(np.isnan(s) for s in params['shape']):
            params['shape'] = tuple(sum(ns) for ns in self.out_nsplits)
        new_op = out.op.copy()
        return new_op.new_tileable(out.inputs, kws=[params])


class TensorIndexHandlerContext(IndexHandlerContext):
    def concat_chunks(self,
                      chunks: List[Chunk],
                      axis: Union[Tuple[int], int]) -> Chunk:
        from ..merge import TensorConcatenate

        assert isinstance(axis, int), \
            'axis to concat could only be int for tensor'

        shape = list(chunks[0].shape)
        shape[axis] = sum(c.shape[axis] for c in chunks)
        chunk_index = list(chunks[0].index)
        chunk_index[axis] = 0

        op = TensorConcatenate(axis=axis, dtype=chunks[0].dtype,
                               sparse=chunks[0].issparse())
        return op.new_chunk(chunks, shape=tuple(shape),
                            index=tuple(chunk_index),
                            order=TensorOrder.C_ORDER)

    def create_chunk(self,
                     chunk_index: Tuple[int],
                     chunk_index_info: ChunkIndexInfo) -> Chunk:
        chunk_op = self.op.copy().reset_key()
        chunk_op._indexes = indexes = chunk_index_info.indexes
        chunk_input = self.tileable.chunks[0] if self.tileable.ndim == 0 else \
            self.tileable.cix[chunk_index]
        chunk_inputs = filter_inputs([chunk_input] + indexes)
        return chunk_op.new_chunk(chunk_inputs,
                                  shape=tuple(chunk_index_info.output_chunk_shape),
                                  index=tuple(chunk_index_info.output_chunk_index),
                                  order=self.op.outputs[0].order)


_type_to_instance = {}


class IndexHandler(ABC):
    @classmethod
    def get_instance(cls):
        if cls not in _type_to_instance:
            _type_to_instance[cls] = cls()
        return _type_to_instance[cls]

    @abstractmethod
    def accept(cls, raw_index):
        pass

    @abstractmethod
    def parse(self,
              raw_index,
              context: IndexHandlerContext) -> IndexInfo:
        pass

    def preprocess(self,
                   index_info: IndexInfo,
                   context: IndexHandlerContext) -> None:
        pass

    @abstractmethod
    def process(self,
                index_info: IndexInfo,
                context: IndexHandlerContext) -> None:
        pass

    def postprocess(self,
                    index_info: IndexInfo,
                    context: IndexHandlerContext) -> None:
        pass

    @classmethod
    def set_chunk_index_info(cls,
                             context: IndexHandlerContext,
                             index_info: IndexInfo,
                             chunk_index: Tuple[int],
                             chunk_index_info: ChunkIndexInfo,
                             output_axis_index: int,
                             index,
                             output_shape: int):
        _ = context, index_info, chunk_index
        chunk_index_info.set(ChunkIndexAxisInfo(output_axis_index=output_axis_index,
                                                processed_index=index,
                                                output_shape=output_shape))


class NewaxisIndexHandler(IndexHandler):
    def accept(self, raw_index):
        return raw_index is np.newaxis

    def parse(self,
              raw_index,
              context: IndexHandlerContext) -> IndexInfo:
        info = IndexInfo(IndexType.new_axis,
                         context.input_axis,
                         context.output_axis,
                         raw_index,
                         self)
        context.output_axis += 1
        context.append(info)
        return info

    def process(self,
                index_info: IndexInfo,
                context: IndexHandlerContext) -> None:
        for chunk_index_info in context.chunk_index_to_info.values():
            # index on axis and index object
            chunk_index_info.set(ChunkIndexAxisInfo(output_axis_index=0,
                                                    processed_index=None,
                                                    output_shape=1))


class SliceIndexHandler(IndexHandler):
    def accept(self, raw_index):
        return isinstance(raw_index, slice)

    def parse(self,
              raw_index,
              context: IndexHandlerContext) -> IndexInfo:
        info = IndexInfo(IndexType.slice,
                         context.input_axis,
                         context.output_axis,
                         raw_index,
                         self)
        context.input_axis += 1
        context.output_axis += 1
        context.append(info)
        return info

    def preprocess(self,
                   index_info: IndexInfo,
                   context: IndexHandlerContext) -> None:
        # make sure input tileable has known chunk shapes
        check_chunks_unknown_shape([context.tileable], TilesError)

    def process(self,
                index_info: IndexInfo,
                context: IndexHandlerContext) -> None:
        tileable = context.tileable
        input_axis = index_info.input_axis
        # slice.step < 0
        is_reversed = (index_info.raw_index.step or 0) < 0

        # e.g. slice_split(slice(3, 10), [2, 2, 7, 5])
        # return {1: slice(1, 2, 1), 2: slice(0, 6, 1)}
        effected_i_to_slice = slice_split(
            index_info.raw_index, tileable.nsplits[index_info.input_axis])
        output_axis_index_range = range(len(effected_i_to_slice)) if not is_reversed else \
            range(len(effected_i_to_slice) - 1, -1, -1)
        other_index_to_iter = dict()

        index_to_info = context.chunk_index_to_info.copy()
        for chunk_index, chunk_index_info in index_to_info.items():
            i = chunk_index[input_axis]
            other_index = chunk_index[:input_axis] + chunk_index[input_axis + 1:]
            size = tileable.nsplits[input_axis][i]
            if i not in effected_i_to_slice:
                # delete it, the input chunk could be ignored
                del context.chunk_index_to_info[chunk_index]
            else:
                slc = effected_i_to_slice[i]
                output_shape = calc_sliced_size(size, slc)
                if other_index not in other_index_to_iter:
                    other_index_to_iter[other_index] = iter(output_axis_index_range)
                output_axis_index = next(other_index_to_iter[other_index])
                self.set_chunk_index_info(context, index_info, chunk_index, chunk_index_info,
                                          output_axis_index, slc, output_shape)


class IntegralIndexHandler(IndexHandler):
    def accept(self, raw_index):
        return isinstance(raw_index, Integral)

    def parse(self,
              raw_index,
              context: IndexHandlerContext) -> IndexInfo:
        info = IndexInfo(IndexType.integer,
                         context.input_axis,
                         context.output_axis,
                         raw_index,
                         self)
        context.input_axis += 1
        context.append(info)
        return info

    def preprocess(self,
                   index_info: IndexInfo,
                   context: IndexHandlerContext) -> None:
        check_chunks_unknown_shape([context.tileable], TilesError)

    def process(self,
                index_info: IndexInfo,
                context: IndexHandlerContext) -> None:
        tileable = context.tileable
        input_axis = index_info.input_axis

        # e.g. slice_split(6, [2, 2, 7, 5])
        # return {2: 2}
        effected_i_to_slice = slice_split(
            index_info.raw_index, tileable.nsplits[index_info.input_axis])

        index_to_info = context.chunk_index_to_info.copy()
        for chunk_index, chunk_index_info in index_to_info.items():
            i = chunk_index[input_axis]
            if i not in effected_i_to_slice:
                # delete it, the input chunk could be ignored
                del context.chunk_index_to_info[chunk_index]
            else:
                slc = effected_i_to_slice[i]
                chunk_index_info.set(ChunkIndexAxisInfo(output_axis_index=None,
                                                        processed_index=slc,
                                                        output_shape=None))


class _BoolIndexHandler(IndexHandler):
    def parse(self,
              raw_index,
              context: IndexHandlerContext) -> IndexInfo:
        info = IndexInfo(IndexType.bool_index,
                         context.input_axis,
                         context.output_axis,
                         raw_index,
                         self)
        context.input_axis += raw_index.ndim
        context.output_axis += 1
        context.append(info)
        return info

    @classmethod
    def _is_first_bool_index(self,
                             context: IndexHandlerContext,
                             index_info: IndexInfo) -> bool:
        bool_index_infos = [info for info in context.parsed_infos
                            if info.index_type == IndexType.bool_index]
        return bool_index_infos[0] is index_info


class NDArrayBoolIndexHandler(_BoolIndexHandler):
    def accept(self, raw_index):
        return isinstance(raw_index, np.ndarray) and \
               raw_index.dtype == np.bool_

    def preprocess(self,
                   index_info: IndexInfo,
                   context: IndexHandlerContext) -> None:
        check_chunks_unknown_shape([context.tileable], TilesError)

    def process(self,
                index_info: IndexInfo,
                context: IndexHandlerContext) -> None:
        tileable = context.tileable
        input_axis = index_info.input_axis
        is_first_bool_index = self._is_first_bool_index(context, index_info)

        axes = list(range(input_axis, input_axis + index_info.raw_index.ndim))
        cum_sizes = []
        for axis in axes:
            cum_sizes.append(np.cumsum((0,) + tileable.nsplits[axis]))

        other_index_to_iter = dict()
        for chunk_index, chunk_index_info in context.chunk_index_to_info.items():
            slcs = []
            for j, axis in enumerate(axes):
                axis_index = chunk_index[axis]
                slcs.append(slice(cum_sizes[j][axis_index],
                                  cum_sizes[j][axis_index + 1]))
            other_index = chunk_index[:axes[0]] + chunk_index[axes[-1] + 1:]
            if other_index not in other_index_to_iter:
                other_index_to_iter[other_index] = itertools.count()
            index = index_info.raw_index[tuple(slcs)]
            output_axis_index = next(other_index_to_iter[other_index])

            # if more than 1 bool index, getitem will rewrite them into fancy
            # but for now, setitem will keep them, thus we cannot record
            # index or shape for this one
            output_axis_index = None if not is_first_bool_index else output_axis_index
            output_size = None if not is_first_bool_index else int(index.sum())

            self.set_chunk_index_info(context, index_info, chunk_index,
                                      chunk_index_info, output_axis_index,
                                      index, output_size)


class TensorBoolIndexHandler(_BoolIndexHandler):
    def accept(self, raw_index):
        return isinstance(raw_index, TENSOR_TYPE) and \
               raw_index.dtype == np.bool_

    def preprocess(self,
                   index_info: IndexInfo,
                   context: IndexHandlerContext) -> None:
        # check both input tileable and index object itself
        check_chunks_unknown_shape([context.tileable, index_info.raw_index], TilesError)

    def process(self,
                index_info: IndexInfo,
                context: IndexHandlerContext) -> None:
        tileable = context.tileable
        input_axis = index_info.input_axis
        index = index_info.raw_index
        # rechunk index into the same chunk size
        nsplits = tileable.nsplits[input_axis: input_axis + index.ndim]
        index = index.rechunk(nsplits)._inplace_tile()
        is_first_bool_index = self._is_first_bool_index(context, index_info)

        other_index_to_iter = dict()
        for chunk_index, chunk_index_info in context.chunk_index_to_info.items():
            effected_chunk_index = chunk_index[input_axis: input_axis + index.ndim]
            other_index = chunk_index[:input_axis] + chunk_index[input_axis + index.ndim:]
            if other_index not in other_index_to_iter:
                other_index_to_iter[other_index] = itertools.count()
            output_axis_index = next(other_index_to_iter[other_index])

            # if more than 1 bool index, getitem will rewrite them into fancy
            # but for now, setitem will keep them, thus we cannot record
            # index or shape for this one
            output_axis_index = None if not is_first_bool_index else output_axis_index
            output_size = None if not is_first_bool_index else np.nan

            self.set_chunk_index_info(context, index_info, chunk_index,
                                      chunk_index_info, output_axis_index,
                                      index.cix[tuple(effected_chunk_index)],
                                      output_size)


class _FancyIndexHandler(IndexHandler):
    def parse(self,
              raw_index,
              context: IndexHandlerContext) -> IndexInfo:
        prev_fancy_indexes = context.get_indexes(IndexType.fancy_index)
        is_first_fancy_index = len(prev_fancy_indexes) == 0

        if is_first_fancy_index:
            output_axis = context.output_axis
        else:
            output_axis = prev_fancy_indexes[0].output_axis
        info = FancyIndexInfo(IndexType.fancy_index,
                              context.input_axis,
                              output_axis,
                              raw_index,
                              self)

        context.input_axis += 1
        if is_first_fancy_index:
            context.output_axis += 1
        context.append(info)
        return info

    @classmethod
    def is_first(cls, index_info: IndexInfo, context: IndexHandlerContext) -> bool:
        # check if is first fancy index after parsing
        fancy_indexes = context.get_indexes(index_info.index_type)
        i = fancy_indexes.index(index_info)
        if i > 0:
            # only process for the first fancy indexes
            return False
        else:
            return True

    def preprocess(self,
                   index_info: IndexInfo,
                   context: IndexHandlerContext) -> None:
        fancy_indexe_infos = context.get_indexes(index_info.index_type)
        # check all fancy indexes are all ndarrays
        for fancy_index_info in fancy_indexe_infos:
            if not self.accept(fancy_index_info.raw_index):  # pragma: no cover
                raise TypeError('Fancy indexes should be all ndarrays or tensors')


class NDArrayFancyIndexHandler(_FancyIndexHandler):
    def accept(self, raw_index):
        return isinstance(raw_index, np.ndarray) and \
               raw_index.dtype != np.bool_

    def preprocess(self,
                   index_info: IndexInfo,
                   context: IndexHandlerContext) -> None:
        is_first = self.is_first(index_info, context)
        if not is_first:
            return

        # check if all ndarrays
        super().preprocess(index_info, context)
        check_chunks_unknown_shape([context.tileable], TilesError)

        fancy_index_infos = context.get_indexes(index_info.index_type)
        # unify shapes of all fancy indexes
        shape = broadcast_shape(*(info.raw_index.shape
                                  for info in fancy_index_infos))
        for fancy_index_info in fancy_index_infos:
            fancy_index_info.shape_unified_index = np.broadcast_to(
                fancy_index_info.raw_index, shape)

        # concat all fancy index together
        concat_fancy_index = np.stack(
            [info.shape_unified_index.ravel() for info in fancy_index_infos])
        effected_nsplits = [context.tileable.nsplits[info.input_axis]
                            for info in fancy_index_infos]
        # split concatenated fancy index into chunks according to input tileable
        split_info = split_indexes_into_chunks(effected_nsplits, concat_fancy_index)
        fancy_index_infos[0].split_info = split_info

    def process(self,
                index_info: IndexInfo,
                context: IndexHandlerContext) -> None:
        fancy_index_infos = context.get_indexes(index_info.index_type)
        fancy_index_axes = [info.input_axis for info in fancy_index_infos]
        split_info = fancy_index_infos[0].split_info
        chunk_index_to_fancy_index_arrays = split_info[0]
        i_fancy_index = fancy_index_infos.index(index_info)

        other_index_to_iter = dict()
        chunk_index_to_info = context.chunk_index_to_info.copy()
        for chunk_index, chunk_index_info in chunk_index_to_info.items():
            effected_chunk_index = tuple(chunk_index[ax] for ax in fancy_index_axes)
            fancy_index_array = \
                chunk_index_to_fancy_index_arrays[effected_chunk_index][i_fancy_index]

            if fancy_index_array.size == 0:
                # not effected
                del context.chunk_index_to_info[chunk_index]
                continue

            if i_fancy_index == 0:
                other_index = tuple(ci for i, ci in enumerate(chunk_index)
                                    if i not in fancy_index_axes)
                if other_index not in other_index_to_iter:
                    other_index_to_iter[other_index] = itertools.count()
                output_axis_index = next(other_index_to_iter[other_index])
                output_axis_shape = fancy_index_array.shape[0]
            else:
                output_axis_index = None
                output_axis_shape = None

            chunk_index_info.set(ChunkIndexAxisInfo(output_axis_index=output_axis_index,
                                                    processed_index=fancy_index_array,
                                                    output_shape=output_axis_shape))

    @classmethod
    def need_postprocess(cls, context: IndexHandlerContext) -> bool:
        fancy_indexes = context.get_indexes(IndexType.fancy_index)

        if fancy_indexes[0].split_info[2] and \
                fancy_indexes[0].shape_unified_index.ndim == 1:
            # if fancy indexes are asc sorted,
            # and they are 1-d, no further computation required
            return False

        return True

    def postprocess(self,
                    index_info: IndexInfo,
                    context: IndexHandlerContext) -> None:
        fancy_indexes = context.get_indexes(index_info.index_type)

        if not self.need_postprocess(context):
            return

        is_first = self.is_first(index_info, context)
        if not is_first:
            # only need to postprocess fancy indexes once
            return

        # current chunks and nsplits
        chunks, nsplits = context.out_chunks, context.out_nsplits

        index_to_chunks = {c.index: c for c in chunks}
        fancy_index_shape = fancy_indexes[0].shape_unified_index.shape
        reorder_index = calc_pos(fancy_index_shape, fancy_indexes[0].split_info[1])

        to_concat_axis = index_info.output_axis
        new_out_chunks = []
        for chunk_index in itertools.product(
                *(range(len(ns)) for ax, ns in enumerate(nsplits)
                  if ax != to_concat_axis)):
            # concat chunks on output axis of first fancy index
            to_concat_chunks = []
            for i in range(len(nsplits[to_concat_axis])):
                to_concat_index = list(chunk_index)
                to_concat_index.insert(to_concat_axis, i)
                to_concat_chunks.append(index_to_chunks[tuple(to_concat_index)])
            concat_chunk = context.concat_chunks(to_concat_chunks, to_concat_axis)

            reorder_chunk_op = context.op.copy().reset_key()
            reorder_chunk_op._indexes = (slice(None),) * to_concat_axis + (reorder_index,)
            reorder_shape = concat_chunk.shape[:to_concat_axis] + fancy_index_shape + \
                concat_chunk.shape[to_concat_axis + 1:]
            chunk_reorder_index = concat_chunk.index[:to_concat_axis] + \
                (0,) * len(fancy_index_shape) + concat_chunk.index[to_concat_axis + 1:]
            reorder_chunk = reorder_chunk_op.new_chunk([concat_chunk],
                                                       shape=reorder_shape,
                                                       index=chunk_reorder_index,
                                                       order=TensorOrder.C_ORDER)
            new_out_chunks.append(reorder_chunk)

        new_nsplits = nsplits[:to_concat_axis] + tuple((s,) for s in fancy_index_shape) \
            + nsplits[to_concat_axis + 1:]
        context.out_chunks = new_out_chunks
        context.out_nsplits = new_nsplits


class TensorFancyIndexHandler(_FancyIndexHandler):
    def accept(self, raw_index):
        return isinstance(raw_index, TENSOR_TYPE) and \
               raw_index.dtype != np.bool_

    def preprocess(self,
                   index_info: IndexInfo,
                   context: IndexHandlerContext) -> None:
        from ..base import broadcast_to
        from ..merge import stack

        is_first = self.is_first(index_info, context)
        if not is_first:
            return

        fancy_index_infos = context.get_indexes(index_info.index_type)

        # check if all tensors
        super().preprocess(index_info, context)
        to_check = [context.tileable] + \
            list(info.raw_index for info in fancy_index_infos)
        check_chunks_unknown_shape(to_check, TilesError)

        # unify shapes of all fancy indexes
        shape = broadcast_shape(
            *(info.raw_index.shape for info in fancy_index_infos))
        fancy_indexes = []
        for fancy_index_info in fancy_index_infos:
            fancy_indexes.append(
                broadcast_to(fancy_index_info.raw_index, shape)._inplace_tile())
        shape_unified_fancy_indexes = unify_chunks(*fancy_indexes)
        for fancy_index_info, shape_unified_fancy_index in \
                zip(fancy_index_infos, shape_unified_fancy_indexes):
            fancy_index_info.shape_unified_index = shape_unified_fancy_index

        fancy_index_axes = tuple(info.input_axis for info in fancy_index_infos)

        # stack fancy indexes into one
        concat_fancy_index = recursive_tile(stack([fancy_index_info.shape_unified_index
                                                   for fancy_index_info in fancy_index_infos]))
        concat_fancy_index = \
            concat_fancy_index.rechunk({0: len(fancy_index_infos)})._inplace_tile()

        self._shuffle_fancy_indexes(concat_fancy_index, context,
                                    index_info, fancy_index_axes)

    @classmethod
    def _shuffle_fancy_indexes(cls,
                               concat_fancy_index: TileableEntity,
                               context: IndexHandlerContext,
                               index_info: IndexInfo,
                               axes: Tuple):
        from .getitem import FancyIndexingDistribute

        tileable = context.tileable

        # generate shuffle map, for concatenated fancy index,
        # calculated a counterpart index chunk for each chunk of input tensor
        map_chunks = []
        for chunk in concat_fancy_index.chunks:
            map_op = FancyIndexingDistribute(
                stage=OperandStage.map, dest_nsplits=tileable.nsplits,
                axes=axes, dtype=chunk.dtype)
            map_chunk = map_op.new_chunk([chunk], shape=(np.nan,),
                                         index=chunk.index,
                                         order=TensorOrder.C_ORDER)
            map_chunks.append(map_chunk)
        # shuffle proxy
        proxy_chunk = TensorShuffleProxy(dtype=concat_fancy_index.dtype).new_chunk(
            map_chunks, shape=(), order=TensorOrder.C_ORDER)
        chunk_index_to_fancy_index_chunks = OrderedDict()
        chunk_index_to_raw_positions = OrderedDict()
        for chunk_index in itertools.product(
                *(range(tileable.chunk_shape[ax]) for ax in axes)):
            shuffle_key = ','.join(str(i) for i in chunk_index)
            reduce_op = FancyIndexingDistribute(
                stage=OperandStage.reduce, axes=axes, dtype=proxy_chunk.dtype,
                shuffle_key=shuffle_key)
            # chunks of fancy indexes on each axis
            kws = [{'axis': ax, 'shape': (np.nan,), 'index': chunk_index,
                    'order': context.op.outputs[0].order}
                   for ax in axes]
            kws.append({'pos': True, 'shape': (np.nan,), 'index': chunk_index})
            reduce_chunks = reduce_op.new_chunks([proxy_chunk], kws=kws)
            chunk_index_to_fancy_index_chunks[chunk_index] = reduce_chunks[:-1]
            chunk_index_to_raw_positions[chunk_index] = reduce_chunks[-1]

        # split info
        #   - chunk_index_to_fancy_index_chunks
        #   - chunk_index_to_raw_positions
        #   - is_fancy_index_asc_sorted, False for tensor fancy indexes
        index_info.split_info = chunk_index_to_fancy_index_chunks, \
            chunk_index_to_raw_positions, False

    def process(self,
                index_info: IndexInfo,
                context: IndexHandlerContext) -> None:
        fancy_index_infos = context.get_indexes(index_info.index_type)
        fancy_index_axes = [info.input_axis for info in fancy_index_infos]
        split_info = fancy_index_infos[0].split_info
        chunk_index_to_fancy_index_chunks = split_info[0]
        i_fancy_index = fancy_index_infos.index(index_info)

        other_index_to_iter = dict()
        for chunk_index, chunk_index_info in context.chunk_index_to_info.items():
            effected_chunk_index = tuple(chunk_index[ax] for ax in fancy_index_axes)
            fancy_index_chunk = \
                chunk_index_to_fancy_index_chunks[effected_chunk_index][i_fancy_index]

            if i_fancy_index == 0:
                other_index = tuple(ci for i, ci in enumerate(chunk_index)
                                    if i not in fancy_index_axes)
                if other_index not in other_index_to_iter:
                    other_index_to_iter[other_index] = itertools.count()
                output_axis_index = next(other_index_to_iter[other_index])
                output_axis_shape = fancy_index_chunk.shape[0]
            else:
                output_axis_index = output_axis_shape = None

            chunk_index_info.set(ChunkIndexAxisInfo(output_axis_index=output_axis_index,
                                                    processed_index=fancy_index_chunk,
                                                    output_shape=output_axis_shape))

    def postprocess(self,
                    index_info: IndexInfo,
                    context: IndexHandlerContext) -> None:
        from .getitem import FancyIndexingConcat

        fancy_index_infos = context.get_indexes(index_info.index_type)

        is_first = self.is_first(index_info, context)
        if not is_first:
            # only need to postprocess fancy indexes once
            return

        # current chunks and nsplits
        chunks, nsplits = context.out_chunks, context.out_nsplits
        chunk_shape = tuple(len(ns) for ns in nsplits)
        to_concat_axis = index_info.output_axis
        tileable = context.tileable
        fancy_index_effected_input_chunk_shapes = tuple(
            tileable.chunk_shape[info.input_axis] for info in fancy_index_infos)
        fancy_indexes = [info.shape_unified_index for info in fancy_index_infos]

        concat_index_to_chunks = dict()
        for chunk in chunks:
            effected_chunk_index = np.unravel_index(
                chunk.index[to_concat_axis], fancy_index_effected_input_chunk_shapes)
            raw_position_chunk = fancy_index_infos[0].split_info[1][effected_chunk_index]
            concat_map_op = FancyIndexingConcat(stage=OperandStage.map,
                                                fancy_index_axis=to_concat_axis,
                                                sparse=chunk.issparse(),
                                                dtype=chunk.dtype)
            map_chunk_shape = \
                chunk.shape[:to_concat_axis] + (np.nan,) + chunk.shape[to_concat_axis + 1:]
            concat_map_chunk = concat_map_op.new_chunk(
                [chunk, raw_position_chunk], index=chunk.index,
                shape=map_chunk_shape, order=TensorOrder.C_ORDER)
            concat_index_to_chunks[concat_map_chunk.index] = concat_map_chunk

        other_index_chunk_shape = chunk_shape[:to_concat_axis] + chunk_shape[to_concat_axis + 1:]
        out_chunks = []
        for chunk_index in itertools.product(*(range(s) for s in other_index_chunk_shape)):
            to_shuffle_chunks = []
            other_shape = None
            for i in range(chunk_shape[to_concat_axis]):
                to_concat_chunk_index = \
                    chunk_index[:to_concat_axis] + (i,) + chunk_index[to_concat_axis:]
                to_concat_chunk = concat_index_to_chunks[to_concat_chunk_index]
                to_shuffle_chunks.append(to_concat_chunk)
                if other_shape is None:
                    other_shape = tuple(s for ax, s in enumerate(to_concat_chunk.shape)
                                        if ax != to_concat_axis)

            proxy_chunk = TensorShuffleProxy(dtype=to_shuffle_chunks[0].dtype).new_chunk(
                to_shuffle_chunks, shape=(), order=TensorOrder.C_ORDER)

            it = itertools.count()
            for reduce_index in itertools.product(
                    *(range(s) for s in fancy_indexes[0].chunk_shape)):
                fancy_index_chunk = fancy_indexes[0].cix[reduce_index]
                concat_reduce_op = FancyIndexingConcat(
                    stage=OperandStage.reduce, fancy_index_axis=to_concat_axis,
                    fancy_index_shape=fancy_index_chunk.shape,
                    dtype=proxy_chunk.dtype, sparse=to_shuffle_chunks[0].issparse(),
                    shuffle_key=str(next(it)))
                reduce_chunk_shape = other_shape[:to_concat_axis] + \
                    fancy_index_chunk.shape + other_shape[to_concat_axis:]
                reduce_chunk_index = chunk_index[:to_concat_axis] + \
                    fancy_index_chunk.index + chunk_index[to_concat_axis:]
                concat_reduce_chunk = concat_reduce_op.new_chunk(
                    [proxy_chunk], shape=reduce_chunk_shape, index=reduce_chunk_index,
                    order=TensorOrder.C_ORDER)
                out_chunks.append(concat_reduce_chunk)

        context.out_chunks = out_chunks
        context.out_nsplits = nsplits[:to_concat_axis] + \
            fancy_indexes[0].nsplits + nsplits[to_concat_axis + 1:]


class IndexesHandler(ABC):
    def __init__(self):
        self.available_index_handlers = []

    def register(self, *handlers):
        self.available_index_handlers.extend(
            h.get_instance() for h in handlers)

    @abstractmethod
    def create_context(self, op):
        pass

    def handle(self, op, return_context: bool = False):
        indexes = op.indexes
        # create context
        context = self.create_context(op)

        # parse index infos
        index_infos = []
        for index in indexes:
            parsed = False
            for index_handler in self.available_index_handlers:
                if index_handler.accept(index):
                    parsed = True
                    index_infos.append(
                        index_handler.parse(index, context))
                    break
            if not parsed:
                raise TypeError(f'unable to parse index {index}')

        self._preprocess(context, index_infos)
        self._process(context, index_infos)
        self._postprocess(context, index_infos)

        if return_context:
            return context
        else:
            return context.create_tileable()

    @classmethod
    def _preprocess(cls,
                    context: IndexHandlerContext,
                    index_infos: List[IndexInfo]):
        # preprocess
        for index_info in index_infos:
            index_info.handler.preprocess(index_info, context)

    @classmethod
    def _process(cls, context, index_infos):
        # process
        for index_info in index_infos:
            index_info.handler.process(index_info, context)

        context.processed_chunks = context.out_chunks = out_chunks = []
        for chunk_index, chunk_index_info in context.chunk_index_to_info.items():
            out_chunks.append(context.create_chunk(chunk_index, chunk_index_info))
        index_to_shape = OrderedDict(sorted([(c.index, c.shape) for c in out_chunks],
                                            key=itemgetter(0)))
        context.out_nsplits = calc_nsplits(index_to_shape)

    @classmethod
    def _postprocess(cls, context, index_infos):
        # post process
        for index_info in index_infos:
            index_info.handler.postprocess(index_info, context)


class NDArrayIndexesHandler(IndexesHandler):
    # indexes handler only for slice, integer,
    # boolean ndarray, integer ndarray and None
    def __init__(self):
        super().__init__()
        self.register(NewaxisIndexHandler,
                      SliceIndexHandler,
                      IntegralIndexHandler,
                      NDArrayBoolIndexHandler,
                      NDArrayFancyIndexHandler)

    def create_context(self, op):
        return TensorIndexHandlerContext(op)

    @classmethod
    def aggregate_result(cls,
                         context: IndexHandlerContext,
                         chunk_index_result: List[Tuple]):
        if len(chunk_index_result) == 1:
            result = chunk_index_result[0][1]
        else:
            result = merge_chunks(chunk_index_result)

        # check if there is any fancy index
        fancy_index_infos = context.get_indexes(IndexType.fancy_index)

        if len(fancy_index_infos) == 0:
            # no fancy index, just return
            return result

        handler = fancy_index_infos[0].handler
        if not handler.need_postprocess(context):
            # no postprocess, just return
            return result

        fancy_index_shape = fancy_index_infos[0].shape_unified_index.shape
        raw_positions = \
            calc_pos(fancy_index_shape, fancy_index_infos[0].split_info[1])
        # reorder the result as the order of fancy indexes
        if isinstance(result, np.ndarray):
            return result[
                (slice(None,),) * fancy_index_infos[0].output_axis + (raw_positions,)]
        else:
            return result.iloc[
                (slice(None,),) * fancy_index_infos[0].output_axis + (raw_positions,)]


class TensorIndexesHandler(IndexesHandler):
    def __init__(self):
        super().__init__()
        self.register(NewaxisIndexHandler,
                      SliceIndexHandler,
                      IntegralIndexHandler,
                      NDArrayBoolIndexHandler,
                      TensorBoolIndexHandler,
                      NDArrayFancyIndexHandler,
                      TensorFancyIndexHandler)

    def create_context(self, op):
        return TensorIndexHandlerContext(op)
